import numpy as np

def InterLin(x,y,p):
    """ 
    Interpolación Lineal.
    Esta función entrega el f(p) a partir de los datos de entrada.
    
    Donde:
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        p: es el punto sobre el cual quiero conocer el valor f(p)
    """
    x = np.array(x);
    y = np.array(y);
    fp = y[0]+ ((y[1]-y[0])*(p-x[0])/(x[1]-x[0]));
    return(fp)
def Inter2(x,y,p):
    """
    Interpolación Cuadrática.
    Esta función entrega el f(p) a partir de los datos de entrada y adoptando
    polinomio de segundo grado.
    
    IMPORTANTE: para poder realizar la interpolación, el tamaño de x y y debe
    ser de 3 datos
    
    Donde:
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        p: es el punto sobre el cual quiero conocer el valor f(p)
    
    
    """
    x = np.array(x);
    y = np.array(y);
    b0 = y[0];
    b1 = (y[1]-y[0])/(x[1]-x[0]);
    b2 = (((y[2]-y[1])/(x[2]-x[1]))-b1)/(x[2]-x[0]);
    fp = b0 + b1*(p-x[0]) + b2*(p-x[0])*(p-x[1]);
    return(fp)
def DifDiv(x,y,p):
    """
    Polinomios de Interpolación de Newton: Diferencias divididas.
    
    Esta función entrega el f(p) a partir de los datos de entrada y se asume 
    como grado de la función interpolante, el que corresponda a n-1 datos de
    entrada.
    
    Donde:
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        p: es el punto sobre el cual quiero conocer el valor f(p) 
    
    """
    x = np.array(x);
    y = np.array(y);
    l = len(x);
    mat = np.zeros([l-1,l-1]);
    vec = [];
    for j in range(0,l-1):
        i=j+1;
        vec.append((y[i]-y[j])/(x[i]-x[j]))
    mat[:,0]=np.transpose(vec);             
         
    c = 2;                       
    for j in range(1,l-1):
        for i in range(l-1-j):
            mat[i,j]=(mat[i+1,j-1]-mat[i,j-1])/(x[i+c]-x[i]);
        c=c+1;
    coef = mat[0,:];
    mat2 = np.ones([len(coef),len(coef)])
    for j in range(len(coef)):
        for i in range(len(coef)-j):
            mat2[i,j] = p-x[i]
    coe=np.prod(mat2,axis=0);
    ccoe = np.zeros([len(coe)])
    for i in range(len(coe)):
        ccoe[i]= coe[len(coe)-1-i]
    fp = np.sum(ccoe*coef)
    fp = fp + y[0];
    return(fp)
def Lagrange(x,y,p,grado):
    """
    Polinomio de Interpolación de Lagrange
    
    Esta función entrega el f(p) a partir de los datos de entrada y el grado
    de la función interpolante.
    
    Donde:
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        p: es el punto sobre el cual quiero conocer el valor f(p)
        grado: grado del polinomio interpolante
    
    """
    x = np.array(x)
    y = np.array(y)
    if(grado>= len(x)):
        print('no se puede obtener polinomio, revise grado de polinomio')
    else:
        mat= np.zeros([grado+1,grado+1])
        for j in range(grado+1):
            for i in range(grado+1):
                if(i==j):
                    mat[i,j] = 1
                else:
                    mat[i,j] = (p-x[i])/(x[j]-x[i])
        coef = np.prod(mat, axis=0)
        ffp = []
        for i in range(grado+1):
           ffp.append(coef[i]*y[i])
        fp = np.sum(ffp)
        return(fp)
def cpoly(x,y,grado):
    """
    Coeficientes de Polinomio interpolante
    
    Esta función entrega los coeficientes en orden ascendente de un polinomio 
    de grado n. (desde a0 hasta an).
    
    y = a0 + a1*x + ... + an*xn
    
    Donde: 
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        grado: grado del polinomio interpolante
        
    NOTA: Debe tener cuidado al usar la función para polinomios de orden 
    superior, pues puede resultar inexactitudes considerables.
        
    """
    x = np.array(x)
    y = np.array(y)
    if(grado>= len(x)):
        print('no se puede obtener polinomio, revise grado de polinomio')
    else:
        mat = np.zeros([grado+1,grado+1])
        for i in range(grado+1):
            for j in range(grado+1):
                mat[i,j] = x[i]**j
        fp = np.matrix(np.linalg.inv(mat))*np.transpose(np.matrix(y))
        return(fp)
def TrazLin(x,y):
    """
    Trazador Lineal
    
    Esta función entrega un conjunto de n ecuaciones de rectas donde n+1 es la
    cantidad de datos ingresados denominadas también trazadores lineales. Los
    datos se entregan en forma de vectores así:
        
        
                        [m0 m1 m2 ... mn], [b0 b1 b2 ... bn]
    
    Donde:
        x: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada       
    
    
    f(x)= m0*x + b0          x0<=x<=x1
    f(x)= m1*x + b1          x1<=x<=x2
    f(x)= m2*x + b2          x1<=x<=x2
    f(x)= mn*x + bn          x(n-1)<=x<=x(n)
    
    NOTA: Cada función sólo es válida en la región comprendida como se ilustra 
    en el conjunto de ecuaciones anteriormente.
    
    """
    x = np.array(x)
    y = np.array(y)
    l=len(x)
    m = []
    b = []
    for i in range(l-1):
        m.append((y[i+1]-y[i])/(x[i+1]-x[i]))
        b.append(y[i]-m[i]*x[i])
    return(m,b)
def Traz2(x,y):
    """
    Trazador cuadrático
    
    Esta función entrega un conjunto de n ecuaciones cuadráticas donde n+1 es 
    la cantidad de datos ingresados denominadas también trazadores cuadráticos. 
    Los datos se entregan en forma de vectores así:
        
             [a0 a1 a2 ... an], [b0 b1 b2 ... bn], [c0 c1 c2 ... cn]
             
    Siendo las constantes a,b,c las constantes de las n-ecuaciones de la forma:
        
        
                f(x)= a*x^2 + b*x + c      xi <= x <= xi+1
                
                
    NOTA: Cada función sólo es válida en la región comprendida como se ilustra 
    anteriormente.                
    
    """
    x = np.array(x)
    y = np.array(y)
    l=len(x)-1
    mat = np.zeros([3*l,3*l])
    vec = np.zeros([3*l-1,1])
    bandera = 1
    ii=0
    jj = 0
    # condición de igualdas en polinomios adyacentes
    for i in range(2*l-2):
        for j in range(3*l):
            bandera = bandera*np.complex(0,1)
            if(np.isreal(bandera)):
                ii = ii
                jj = jj +3
            else:
                ii=ii+1
                jj = jj
            mat[i,jj]=x[ii]**2
            mat[i,jj+1]=x[ii]
            mat[i,jj+2]=1
            vec[i,0]=y[ii]
            break
    #condiciones de bordes externos          
    mat[2*l-2,0]=x[0]**2
    mat[2*l-2,1]=x[0]
    mat[2*l-2,2]=1
    vec[2*l-2,0]=y[0]
    mat[2*l-1,-3]=x[-1]**2
    mat[2*l-1,-2]=x[-1]
    mat[2*l-1,-1]=1
    vec[2*l-1,0]=y[-1]
    #condición de igualdad en primeras derivadas de polinomios adyacentes
    bandera = 1
    ii=0
    jj = 0   
    for i in range(l-1):
        for j in range(3*l):
            ii = range(2*l,3*l-1)
            bandera=  bandera*np.complex(0,1)
            if(np.isreal(bandera)):
                jj = jj +3
            else:
                jj = jj
            mat[ii[i],jj]=2*x[i+1]
            mat[ii[i],jj+1]=1
            mat[ii[i],jj+3]=-2*x[i+1]
            mat[ii[i],jj+4]=-1
            break
    #condición de segunda derivada igual a 0 en punto inicial
    mat[3*l-1,0]=1
    mat2=np.zeros([3*l-1,3*l-1])  
    for i in range(3*l-1):
        for j in range(3*l-1):
            mat2[i,j]=mat[i,j+1]
          
    #encontrando los valores para los an's
    X = np.matrix(np.linalg.inv(mat2))*np.matrix(vec)
    a = [0]
    b=[]
    c=[]
    co=0
    for i in range(3):
        a.append(np.double(X[i*3-co]))
        b.append(np.double(X[i*3]))
        c.append(np.double(X[i*3+1]))
        co=1
    return(a,b,c)
    
            
                    


import numpy as np

def RegLin(t,y):
    """
    Regresión Lineal
    
    Ajusta los puntos ingresados (t,y) a una línea recta bajo el críterio de 
    mínimos cuadrados
    
    Donde:
        t: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
    
    y = m*x +b
    
    La función entrega 2 valores, los cuales son m y b de la ecuación anterior
    respectivamente.
    """
    t = np.array(t);
    y = np.array(y);
    l = len(t);
    sum1 = np.sum(t*y);
    sum2 = np.sum(t**2);
    tmed = np.sum(t)/l;
    ymed = np.sum(y)/l;
    a1 = ((l*sum1)-(np.sum(t)*np.sum(y)))/((l*sum2)-(np.sum(t))**2)
    a0 = ymed - a1*tmed
    return(a1,a0)

def RegPol(t,y,grado):  # se resuelve con n ecuaciones para n incognitas
    """
    Regresión Polinomial
    
    Ajusta los puntos ingresados (t,y) a un polinomio de un grado cualquiera,
    siempre y cuando la cantidad de datos sea mayor al grado + 1 deseado.
    
    Donde:
        t: es el vector de datos de la absisa
        y: es el vector de datos de la ordenada
        grado: es el grado del polinomio al que se quiere ajustar los datos
        
    La función entrega n valores, los cuales son los coeficientes del polinomio
    solicitado. tenga en cuenta que el orden de los coeficientes es desde el
    valor independiente hacia el coeficiente de la variable de mayor grado.
    
    y = a0 + a1*x + a2*x^2 + ... + an*x^n
    
    """
    t = np.array(t);
    y = np.array(y);
    l = len(t);
    grado = grado +1;
    if(l<=grado):
        print('no es posible realizar regresión.')
        print('Revise el la cantidad de datos y el orden del polinomio deseado')
    else:
        mat = np.zeros([grado,grado])
        for i in range(grado):
            for j in range(grado):
                mat[i,j]=np.sum(t**(j+i))
        mat[0,0]=l
    vec = np.zeros([grado,1])
    for i in range(grado):
        vec[i,0]=np.sum((t**i)*(y))
           
    A = np.matrix(np.linalg.inv(mat))*np.matrix(vec);
    return(A)
  
def RegLinMul(t,y):
    """
    Regresión Lineal Multiple
    
    Ajusta los datos ingresados a un polinomio de igual cantidad de variables
    a la cantidad de vectores de datos que posea t.
    
    Donde:
        t: es una matriz con los datos experimentales
        y: vector de datos depentientes de t
    
    Ejemplo:
        Ajuste los siguientes datos:
                x1   X2    y
                0    0     5
                2    1     10
                2.5  2     9
                1    3     0
                4    6     3
                7    2     27
        dado que los datos dependen de 2 variables, el algoritmo tendrá que 
        encontrar valores para los coeficientes de una ecuación que depende de
        2 variables. Esto se hace por defecto.
        
        Ingreso de datos:
            t = [[0,2,2.5,1,4,7],
                 [0,1,2,3,6,2]]
            y = [5,10,9,0,3,27],
            
        Salida del algoritmo:
            La función entregará un vector con los coeficientes de la ecuación
            que tiene la forma siguiente:
            
                    y = a0 + a1*x1 + a2*x2 + ... + an*xn
        
    
    """
    t = np.array(t);
    y = np.array(y);
    l = len(t)+1;
    mat = np.zeros([l,l]);
    for i in range(l-1):
        for j in range(l-1):
            if(i==j):
                mat[i+1,j+1]=np.sum(t[i]**2);
            else:
                mat[i+1,j+1]=np.sum(t[i]*t[j]);
            if(i==0 and j==0):
                mat[i,j]=len(t[0]);
            if(i==0):
                mat[i,j+1]=np.sum(t[j]);
            if(j==0):
                mat[i+1,j]=np.sum(t[i]);
    vec=np.zeros([l,1]);
    for i in range(l-1):
        vec[i+1,0]=np.sum(t[i]*y);
    vec[0,0] = np.sum(y);
    x = np.matrix(np.linalg.inv(mat))*np.matrix(vec)
    return(x)
    
    

           


import numpy as np
def cramer(A,B):
    """
    Esta función permite resolver sistemas lineales de n ecuaciones 
    con n incognitas utilizando la regla de cramer.
    
                        [A]*[X]=[B]
    
    donde:
        A: es la matriz de coeficientes
        B: es el vector de términos independientes
    
    La función entrega el vector de soluciones para X
    
    """
    B=np.matrix(B)
    D = np.linalg.det(A)
    X = []
    for i in range(len(A)):
        AA = np.matrix(A)
        for j in range(len(A)):
            AA[j,i] = B[j,0]
        DD = np.linalg.det(AA)
        X.append(DD/D)
    return(X)
def desc_LU(A):
    """
    Descomposición LU -> matrices "L" "U"
    
    Esta función devuelve las matrices "L" y "U" a partir de una matriz de 
    coeficientes A.
    """
    mat=np.matrix(A)
    n = len(mat)
    L = np.matrix(np.zeros([n,n]))
    for j in range(0,n-1):
        for i in range(1+j,n):
            fac = (-mat[i,j]/mat[j,j])
            L[i,j]=-fac
            for q in range(0,n):
                mat[i,q]= (fac*mat[j,q]+mat[i,q])
    U = mat
    for i in range(n):
        for j in range(n):
            if i == j:
                L[i,j]=1
    return(L,U)
def desc_D(L,b):
    """
    descomposición LU -> matriz "D"
    
    Esta función devuelve la matriz "D" a partir de la matriz "L" (que se puede
    obtener con la función "desc_LU") y el vector de términos independientes.
    
    
    """
    L = np.matrix(L)
    b = np.matrix(b)
    D = (np.linalg.inv(L)*b)
    return(D)



import numpy as np

def Trap(x,y,h='0'):
    """
    Regla del Trapecio
    
    Esta función permite calcular la integral numérica de los datos (x,y). El 
    valor de h es opcional y si se usa, no es necesario tener los datos de x.
    
    la función retorna el valor de la integral por la regla del trapecio.
    
    NOTA: Si se poseen los datos (x,y), estos no necesitan estar igualmente
          para calcular la integral
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    II=[]
    h = np.double(h)
    if(h==0):
        for i in range(l-1):
            II.append((x[i+1]-x[i])*(y[i]+y[i+1])/2)
        I = np.sum(II)            
    else:
        suma = 0
        for i in range(1,l-1):
            suma = suma + y[i]
            suma = 2*suma
        I = h*(y[0]+ suma + y[-1])/2
    return(I)
def Simpson(x,y):
    """
    Regla Simpson 1/3
    
    Esta función permite calcular la integral numérica de los datos (x,y).
    
    la función retorna el valor de la integral por la regla de simpson 1/3.
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    h = (x[-1]-x[0])
    suma1 =0
    suma2=0
    for i in range(1,l-1,2):
        suma1 = suma1+ 4*y[i]
    for i in range(2,l-1,2):
        suma2 = suma2+ 2*y[i]
    I = h*(y[0]+suma1+suma2+y[-1])/(3*(l-1))
    return(I)
def Simpson3(x,y):
    """
    Regla Simpson 3/8
    
    Esta función permite calcular la integral numérica de los datos (x,y). La 
    función retorna el valor de la integral por la regla de simpson 1/3.
    
    NOTA: esta función solo calcula la integral bajo la curva denotada por los
    4 puntos, es decir no realiza el cálculo de la integral definida por mas 
    puntos.
    
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    h = x[-1]-x[0]
    I = h*(y[0]+3*(y[1]+y[2])+y[3])/8
    return(I)
    
    
    


x = [0,0.25,0.5,0.75,1]
y = [1.2,1.103516,0.925,0.6363281,0.2]

import numpy as np
def dif(x,y,p,h,HP='1'):
    """
    Diferencia Dividida Finita
    
    Esta función entrega la derivada de los datos (x,y) evaluados en el 
    punto P dado. Las formulas empleadas para calcular la derivada corresponden
    a la diferencia dividida finita centrada, con lo cual es necesario que el
    punto P tenga al menos un valor hacia atrás y otro hacia adelante para la
    versión simplificada (HP=0) o 2 valores hacia atrás y 2 hacia adelante para
    la versión de alta presición (HP=1)
    
    NOTAS: 
        Para definir el punto p, se utiliza la posición del dato en la cual
        se desea evaluar la derivada.
        El la diferencia entre X[i+1]-X[i] debe ser constante
    
    Ejemplo: Evaluar la derivada en x=3 de los siguientes datos
        x = [2,3,4], y=[8,12,16]
        - Posición de interés para evaluar la derivada en 1 porque x[1]=3.
        - Dado que a aprtir de la posición en la que se evaluará la derivada 
          solo tengo un dato hacia atrás y uno hacia adelante, se debe 
          selecionar baja presición (HP=0).
        - El paso (h) es x[1]-x[0]=1 para este caso
        
        
                              x = [2,3,4]
                              y=[8,12,16]
                              dif(x,y,1,1,0)
                              >> 4.0
        
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    p = np.int(p)
    HP = np.int(HP)
    if(HP==1):
        if(p>=2):
            D = (-y[p+2]+8*y[p+1]-8*y[p-1]+y[p-2])/(12*h)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela o trabaje con fórmula '
                  'menos precisa (HP=0)')
    else:
        if(p>=1):
            D = (y[p+1]-y[p-1])/(2*h)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela')
    return(D)
def dif2(x,y,p,h,HP='1'):
    """
    Diferencia Dividida Finita
    
    Esta función entrega la segunda derivada de los datos (x,y) evaluados en el 
    punto P dado. Las formulas empleadas para calcular la derivada corresponden
    a la diferencia dividida finita centrada, con lo cual es necesario que el
    punto P tenga al menos un valor hacia atrás y otro hacia adelante para la
    versión simplificada (HP=0) o 2 valores hacia atrás y 2 hacia adelante para
    la versión de alta presición (HP=1)
    
    NOTAS: 
        Para definir el punto p, se utiliza la posición del dato en la cual
        se desea evaluar la derivada.
        El la diferencia entre X[i+1]-X[i] debe ser constante
    
    Ejemplo: Remítase a la ayuda de la función "dif"
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    p = np.int(p)
    HP = np.int(HP)
    if(HP==1):
        if(p>=2):
            D=(-y[p+2]+16*y[p+1]-30*y[p]+16*y[p-1]-y[p-2])/(12*h**2)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela o trabaje con fórmula '
                  'menos precisa (HP=0)')
    else:
        if(p>=1):
            D=(y[p+1]-2*y[p]+y[p-1])/(h**2)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela')
    return(D)
def dif3(x,y,p,h,HP='1'):
    """
    Diferencia Dividida Finita
    
    Esta función entrega la tercera derivada de los datos (x,y) evaluados en el 
    punto P dado. Las formulas empleadas para calcular la derivada corresponden
    a la diferencia dividida finita centrada, con lo cual es necesario que el
    punto P tenga al menos 2 valores hacia atrás y 2 hacia adelante para la
    versión simplificada (HP=0) o 3 valores hacia atrás y 3 hacia adelante para
    la versión de alta presición (HP=1)
    
    NOTAS: 
        Para definir el punto p, se utiliza la posición del dato en la cual
        se desea evaluar la derivada.
        El la diferencia entre X[i+1]-X[i] debe ser constante
    
    Ejemplo: Remítase a la ayuda de la función "dif"
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    p = np.int(p)
    HP = np.int(HP)
    if(HP==1):
        if(p>=3):
            D=(-y[p+3]+8*y[p+2]-13*y[p+1]+13*y[p-1]-8*y[p-2]+y[p-3])/(8*h**3)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela o trabaje con fórmula '
                  'menos precisa (HP=0)')
    else:
        if(p>=2):
            D=(y[p+2]-2*y[p+1]+2*y[p-1]-y[p-2])/(2*h**3)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela')
    return(D)
def dif4(x,y,p,h,HP='1'):
    """
    Diferencia Dividida Finita
    
    Esta función entrega la tercera derivada de los datos (x,y) evaluados en el 
    punto P dado. Las formulas empleadas para calcular la derivada corresponden
    a la diferencia dividida finita centrada, con lo cual es necesario que el
    punto P tenga al menos 2 valores hacia atrás y 2 hacia adelante para la
    versión simplificada (HP=0) o 3 valores hacia atrás y 3 hacia adelante para
    la versión de alta presición (HP=1)
    
    NOTAS: 
        Para definir el punto p, se utiliza la posición del dato en la cual
        se desea evaluar la derivada.
        El la diferencia entre X[i+1]-X[i] debe ser constante
    
    Ejemplo: Remítase a la ayuda de la función "dif"
    
    """
    x = np.array(x)
    y = np.array(y)
    l = len(y)
    p = np.int(p)
    HP = np.int(HP)
    if(HP==1):
        if(p>=3):
            D=(-y[p+3]+12*y[p+2]+39*y[p+1]+56*y[p]-39*y[p-1]+12*y[p-2]+y[p-3])/(6*h**4)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela o trabaje con fórmula '
                  'menos precisa (HP=0)')
    else:
        if(p>=2):
            D=(y[p+2]-4*y[p+1]+6*y[p]-4*y[p-1]+y[p-2])/(h**4)
        else:
            print('NO es posible evaluar la derivada en la posición '
                  'seleccionada, por favor verifíquela')
    return(D)
        
    
            
        
